#include <stdlib.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <packet.h>
#include <cleanup.h>
#include <capdev.h>
#include <capture.h>
#include <decode.h>
#include <preproc.h>
#include <plugin.h>

#include <pcap.h>

#define READ_TIMEOUT 500

struct fpcap_priv {
	pcap_t		*pcap_desc;
	struct packet	pkt;
	struct proto	*proto;
};

proc_decode_subproto decode_subproto;
proc_preproc_dispatch preproc_dispatch;
proc_serial_number serial_number;

void *fpcap_init(char *interface)
{
	unsigned int lnk;
	struct fpcap_priv *priv;
	char ebuf[PCAP_ERRBUF_SIZE];

	if ( !(priv=calloc(1, sizeof(struct fpcap_priv))) )
		return NULL;

	if ( !(priv->pcap_desc=pcap_open_live(interface, 4096,
		0, /* Not promiscous */
		READ_TIMEOUT, ebuf)) ) {
		printf("pcap: %s\n", ebuf);
		free(priv);
		return NULL;
	}

	lnk=pcap_datalink(priv->pcap_desc);
	
	if ( !(priv->proto=decode_subproto("__pcap_dlt",lnk)) ) {
		printf("pcap: %s: Can't support protocol 0x%x\n", interface, lnk);
		pcap_close(priv->pcap_desc);
		free(priv);
		return NULL;
	}

	priv->pkt.flags=FP_CLONE|FP_LIVE;
	
	return priv;
}

void fpcap_end(void *priv)
{
	struct fpcap_priv *p=priv;

	if ( p && p->pcap_desc ) pcap_close(p->pcap_desc);
	if ( p ) free(p);
}

void lpf_callback(u_char *user, struct pcap_pkthdr *header, u_char *data)
{
	struct fpcap_priv *p;

	if ( !(p=(struct fpcap_priv *)user) ) return;

	/* Pointers */
	p->pkt.base=data;
	p->pkt.end=data;
	p->pkt.end+=header->caplen;

	/* length */	
	p->pkt.caplen=header->caplen;
	p->pkt.len=header->len;
	
	/* Timestamp */
	p->pkt.time.tv_sec=header->ts.tv_sec;
	p->pkt.time.tv_usec=header->ts.tv_usec;

	/* first layer */
	p->pkt.layer[0].proto=p->proto;
	p->pkt.layer[0].h.raw=p->pkt.base;
	p->pkt.layer[0].flags=0;
	p->pkt.llen=1;
	p->pkt.present=0;
	p->pkt.rlen=0;

	if ( p->proto->decode(&p->pkt) ) {
		preproc_dispatch(&p->pkt);
	}
}

void fpcap_go(void *priv, struct cap_s *c)
{
	struct fpcap_priv *p=priv;

	p->pkt.capture=c;

	while(c->state!=CAP_STATE_STOP) {
		serial_number(&p->pkt.serial);
		if ( pcap_dispatch(p->pcap_desc, 1,
			(pcap_handler)lpf_callback, priv)<0 ) break;
	}
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	int ok=0;
	proc_capdev_add capdev_add;

	/* validate input */
	plugin_check(in, out);

	/* tell firestorm who we are */
	PLUGIN_ID("capture.pcap", "Live libpcap capture");
	PLUGIN_VERSION(1, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");

	if ( !(capdev_add=in->import("capdev.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(decode_subproto=in->import("decode.subproto")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(preproc_dispatch=in->import("preproc.dispatch")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(serial_number=in->import("serial_number")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	ok+=capdev_add("pcap", fpcap_init, fpcap_end, fpcap_go);
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL ;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
