#include "tcpip.h"

/* Realm stuff */
struct realm *ip_realm;

/* Rule header list for IP */
struct ipm_head *iprules=NULL;

unsigned int num_ip_head=0;
unsigned int num_ip_rules=0;

/* Decode Stuff */
struct proto ipv4_p=init_proto_ex("ipv4", ipv4_decode, ipv4_dprint);
struct proto_req ipv4_r[]={
	proto_request("ethernet", __constant_htons(0x0800)),
	proto_request("sll", __constant_htons(0x0800)),
	proto_request("linux", __constant_htons(0x0800)),
	proto_request("__pcap_dlt", 0),
	proto_request("__pcap_dlt", 12), /* DLT_RAW */
	null_request()
};

/* Match a packet to our rulesets */
void ip_match(struct packet *p, unsigned int l)
{
	struct ipm_head *h;
	struct pkt_iphdr *iph;

	iph=p->layer[l].h.ip;

	for(h=iprules; h; h=h->next) {
		if (
			/* Source addr  */
			(((h->saddr&h->smask) == (iph->saddr&h->smask))
			^ ((h->flags&NEGATE_SADDR)!=0) )&&
			
			/* Destination address */
			(((h->daddr&h->dmask) == (iph->daddr&h->dmask))
			^ ((h->flags&NEGATE_DADDR)!=0)) ) {
			
			/* Check Protocol */
			if ( h->flags&CHECK_PROTO &&
				!((h->proto == iph->protocol)
				^ ((h->flags&NEGATE_PROTO)!=0)) ) continue;

			if(tcpip_traverse("sig.ip", p, l, h->rules))
				return;
		}
	}

	/* Don't bother if there isn't any more
	 * decode information. This could happen
	 * because this is a fragment or something */
	if ( l+1 >= p->llen ) return;
	
	/* In order of likelieness */
	if ( iph->protocol==6 )  {tcp_match(p, l);  return;}
	if ( iph->protocol==17 ) {udp_match(p, l);  return;}
	if ( iph->protocol==1 )  {icmp_match(p, l); return;}
}

struct ipm_head *ip_rlink(struct ipm_head *n)
{
	struct ipm_head *tmp;

	for(tmp=iprules; tmp; tmp=tmp->next) {
		if ( n->saddr==tmp->saddr && n->smask==tmp->smask &&
		  	n->daddr==tmp->daddr && n->dmask==tmp->dmask &&
			n->proto==tmp->proto && n->flags==tmp->flags ) {
			/* We found an identical node, free
			 * the shit and return the node from
			 * the list */
			free(n);
			return tmp;
		}
	}

	/* Link this node in with the list */
	num_ip_head++;
	n->next=iprules;
	iprules=n;
	return n;
}

/* The signature subsystem commits rules to us here.
 * We must copy out all the relevent data to our internal
 * structures. the rule and all the criterion get freed
 * after we return. The alert data is already setup and
 * we can just copy the pointers - or cleanup if we must */
int ip_commit(struct rule *r)
{
	struct ip_rule *rule;
	struct ipm_head *hd;
	struct criteria *t;
	int err=0;

	/* Allocate a rule header */
	if ( !(hd=calloc(1, sizeof(*hd))) ) {
		return 0;
	}

	/* Allocate a rule */
	if ( !(rule=calloc(1, sizeof(*rule))) ) {
		free(hd);
		return 0;
	}
	
	for(t=r->criteria; t; t=t->next) {
		if ( !strcmp("src", t->crit) ) {
			if ( t->negate ) hd->flags|=NEGATE_SADDR;
			if ( !parse_ipaddr(t->args, &hd->saddr, &hd->smask) ) {
				err=1;
				break;
			}
			continue;
		}

		if ( !strcmp("dst", t->crit) ) {
			if ( t->negate ) hd->flags|=NEGATE_DADDR;
			if ( !parse_ipaddr(t->args, &hd->daddr, &hd->dmask) ) {
				err=1;
				break;
			}
			continue;
		}

		/* Ports */
		if ( !strcmp(t->crit, "sport") || !strcmp(t->crit, "dport") ) {
			printf("Can't handle ports in IP realm! (yet)\n");
			err=1;
			break;
		}

		if ( !strcmp("ip_proto", t->crit) ) {
			if ( t->negate ) hd->flags|=NEGATE_PROTO;
			hd->flags|=CHECK_PROTO;
			if ( !parse_ipproto(t->args, &hd->proto) ) {
				err=1;
				break;
			}
			continue;
		}

		if ( !iprule_addcrit(t, rule, r->modifier) ) {
			err=1;
			break;
		}
	}

	if ( err ) {
		iprule_cleanup(rule);
		free(hd);
		return 0;
	}

	/* Can't fail */
	hd=ip_rlink(hd);

	/* Setup rule structures */
	memcpy(&rule->a, &r->alert, sizeof(rule->a));

	/* Link rule on to rule header */
	num_ip_rules++;
	if ( !hd->rules ) {
		hd->rules=rule;
		hd->rules_tail=rule;
	}else{
		hd->rules_tail->next=rule;
		hd->rules_tail=rule;
	}

	return 1;
}

/* Print out an IPv4 header */
int ipv4_dprint(struct layer *l, char *buf, int buflen)
{
	struct in_addr s, d;
	char sip[32], dip[32];

	s.s_addr=l->h.ip->saddr;
	d.s_addr=l->h.ip->daddr;
	strncpy(sip, inet_ntoa(s), 32);
	strncpy(dip, inet_ntoa(d), 32);
	
	return snprintf(buf, buflen,
		"%s -> %s TTL=%u ID=%u PROTO=%u %u@%u",
		sip,
		dip,
		l->h.ip->ttl,
		l->h.ip->id,
		l->h.ip->protocol,
		ntohs(l->h.ip->tot_len),
		(ntohs(l->h.ip->frag_off)&IP_OFFMASK)<<3);
}

static inline unsigned int ip_csum(struct pkt_iphdr *iph)
{
	u_int16_t *tmp=(u_int16_t *)iph;
	u_int16_t csum;
	u_int32_t sum=0;
	int i;
	
	for(i=0; i<iph->ihl<<1; i++) {
		/* ignore field 5, its the checksum ;) */
		if ( i!=5 ) sum+=tmp[i];
		if(sum & 0x80000000) {
			sum = (sum & 0xffff) + (sum >> 16);
		}
	}

	while(sum >> 16)
		sum = (sum & 0xffff) + (sum >> 16);
	
	csum=(~sum) & 0xffff;
	
	if ( csum != iph->csum) {
		printf("fuck: %u %u\n", csum, iph->csum);
		return 0;
	}

	return FLAG_IP_CSUM;
}

/* Decode an IPv4 packet */
int ipv4_decode(struct packet *p)
{
	struct proto_child *pc;
	struct layer *l=&p->layer[p->llen-1];
	int ret=0;

	if ( p->llen >= PKT_LAYERS ) return ret;

	/* Check it meets minimum size of packet */
	if ( l->h.raw+sizeof(struct pkt_iphdr) >= p->end )
		return ret;

	/* Check the header length isn't bogus, encapsulated
	 * protocols can't overlap the header! Also check the
	 * version field, it can only be IPv4 */
	if ( l->h.ip->ihl<5 || l->h.ip->version!=4 ) {
		static struct alert a={alert:"tcpdump ihl exploit"};
		alert("decode.ip", p, &a);
		return ret;
	}

	/* We don't really need to check the checksum because
	 * you can't use bad checksums to exploit anything, no
	 * routers will actually pass them on. Though you could
	 * generate attacks locally or it could be a genuine error.
	 * either way, its only 20 bytes! :) */
	if ( !(l->flags&FLAG_IP_CSUM) )
		l->flags|=ip_csum(l->h.ip);

	/* Use the ihl header field to decode the packet */
	if ( (p->layer[p->llen].h.raw=
		l->h.raw+(l->h.ip->ihl<<2))
		>= p->end ) return ret;

	/* Simple example of a realm. All ipv4
	 * packets are in the "ip" realm */
	if ( ip_realm ) {
		p->realm[p->rlen].layer=p->llen-1;
		p->realm[p->rlen++].realm=ip_realm;
		ret=1;
	}
	
	/* Only decode fragments with offset 0, ie:
	 * those containing the next header */
	if ( l->h.ip->frag_off & __constant_htons(IP_OFFMASK) )
		goto finish;
	
	/* Find a relevent child */
	p->layer[p->llen].flags=0;
	for(pc=l->proto->children; pc; pc=pc->next)
	{
		if ( l->h.ip->protocol == pc->id ) {
			p->layer[p->llen++].proto=pc->proto;
			return pc->proto->decode(p) | ret;
		}
	}

finish:
	p->layer[p->llen++].proto=NULL;
	return ret;
}
